@@ -1,13 +1,13 @@
 # -*- test-case-name: wokkel.test.test_xmppim -*-
 #
-# Copyright (c) 2003-2009 Ralph Meijer
+# Copyright (c) Ralph Meijer.
 # See LICENSE for details.
 
 """
 XMPP IM protocol support.
 
 This module provides generic implementations for the protocols defined in
-U{RFC 3921<http://www.xmpp.org/rfcs/rfc3921.html>} (XMPP IM).
+U{RFC 3921<http://xmpp.org/rfcs/rfc3921.html>} (XMPP IM).
 
 All of it should eventually move to Twisted.
 """
@@ -296,6 +296,18 @@ class AvailabilityPresence(BasePresence):
         self.priority = priority
 
 
+    def __get_status(self):
+        if None in self.statuses:
+            return self.statuses[None]
+        elif self.statuses:
+            for status in self.status.itervalues():
+                return status
+        else:
+            return None
+
+    status = property(__get_status)
+
+
     def _childParser_show(self, element):
         show = unicode(element)
         if show in ('chat', 'away', 'xa', 'dnd'):
@@ -350,7 +362,7 @@ class SubscriptionPresence(BasePresence):
     This kind of presence is used to represent requests for presence
     subscription and their replies.
 
-    Based on L{BasePresence} and {Stanza}, it just uses the L{stanzaType}
+    Based on L{BasePresence} and {Stanza}, it just uses the C{stanzaType}
     attribute to represent the type of subscription presence. This can be
     one of C{'subscribe'}, C{'unsubscribe'}, C{'subscribed'} and
     C{'unsubscribed'}.
@@ -367,31 +379,32 @@ class ProbePresence(BasePresence):
 
 
 
-class PresenceProtocol(XMPPHandler):
+class BasePresenceProtocol(XMPPHandler):
     """
-    XMPP Presence protocol.
+    XMPP Presence base protocol handler.
+
+    This class is the base for protocol handlers that receive presence
+    stanzas. Listening to all incoming presence stanzas, it extracts the
+    stanza's type and looks up a matching stanza parser and calls the
+    associated method. The method's name is the type + C{Received}. E.g.
+    C{availableReceived}. See L{PresenceProtocol} for a complete example.
 
     @cvar presenceTypeParserMap: Maps presence stanza types to their respective
         stanza parser classes (derived from L{Stanza}).
     @type presenceTypeParserMap: C{dict}
     """
 
-    presenceTypeParserMap = {
-                'error': ErrorStanza,
-                'available': AvailabilityPresence,
-                'unavailable': AvailabilityPresence,
-                'subscribe': SubscriptionPresence,
-                'unsubscribe': SubscriptionPresence,
-                'subscribed': SubscriptionPresence,
-                'unsubscribed': SubscriptionPresence,
-                'probe': ProbePresence,
-        }
+    presenceTypeParserMap = {}
 
     def connectionInitialized(self):
         self.xmlstream.addObserver("/presence", self._onPresence)
 
 
+
     def _onPresence(self, element):
+        """
+        Called when a presence stanza has been received.
+        """
         stanza = Stanza.fromElement(element)
 
         presenceType = stanza.stanzaType or 'available'
@@ -411,6 +424,21 @@ class PresenceProtocol(XMPPHandler):
             handler(presence)
 
 
+
+class PresenceProtocol(BasePresenceProtocol):
+
+    presenceTypeParserMap = {
+                'error': ErrorStanza,
+                'available': AvailabilityPresence,
+                'unavailable': AvailabilityPresence,
+                'subscribe': SubscriptionPresence,
+                'unsubscribe': SubscriptionPresence,
+                'subscribed': SubscriptionPresence,
+                'unsubscribed': SubscriptionPresence,
+                'probe': ProbePresence,
+                }
+
+
     def errorReceived(self, presence):
         """
         Error presence was received.
@@ -699,6 +727,46 @@ class RosterClientProtocol(XMPPHandler):
         @type entity: L{JID}
         """
 
+
+
+class Message(Stanza):
+    """
+    A message stanza.
+    """
+
+    stanzaKind = 'message'
+
+    childParsers = {
+            (None, 'body'): '_childParser_body',
+            (None, 'subject'): '_childParser_subject',
+            }
+
+    def __init__(self, recipient=None, sender=None, body=None, subject=None):
+        Stanza.__init__(self, recipient, sender)
+        self.body = body
+        self.subject = subject
+
+
+    def _childParser_body(self, element):
+        self.body = unicode(element)
+
+
+    def _childParser_subject(self, element):
+        self.subject = unicode(element)
+
+
+    def toElement(self):
+        element = Stanza.toElement(self)
+
+        if self.body:
+            element.addElement('body', content=self.body)
+        if self.subject:
+            element.addElement('subject', content=self.subject)
+
+        return element
+
+
+
 class MessageProtocol(XMPPHandler):
     """
     Generic XMPP subprotocol handler for incoming message stanzas.
